/*
 * Dump R4x00 TLB for debugging purposes.
 *
 * Copyright (C) 1994, 1995 by Waldorf Electronics,
 * written by Ralf Baechle.
 */

#include <linux/kernel.h>
#include <linux/string.h>

#include <asm/bootinfo.h>
#include <asm/cachectl.h>
#include <asm/mipsregs.h>

void dump_tlb(void)
{
	int	i;
	int	wired;
	unsigned int pagemask;
	unsigned long long entryhi, entrylo0, entrylo1;

	wired = read_32bit_cp0_register(CP0_WIRED);
	printk("Wired: %d", wired);
	
	for(i=0;i<boot_info.tlb_entries;i++)
	{
		write_32bit_cp0_register(CP0_INDEX, i);
		__asm__ __volatile__(".set noreorder;nop;tlbr;nop;nop;nop;.set reorder");

		pagemask = read_32bit_cp0_register(CP0_PAGEMASK);
		entryhi  = read_64bit_cp0_register(CP0_ENTRYHI);
		entrylo0 = read_64bit_cp0_register(CP0_ENTRYLO0);
		entrylo1 = read_64bit_cp0_register(CP0_ENTRYLO1);

		/* if((entrylo0|entrylo1) & 2) */
		/* if(1) */
		if((entryhi >> 32) & 3 != 2)
		{
			/*
			 * Only print entries in use
			 */
			printk("\nIndex: %2d  %08x", i, pagemask);

			printk("  %08x %08x", (unsigned int)(entryhi >> 32),
			                      (unsigned int) entryhi);
			printk("  %08x %08x", (unsigned int)(entrylo0 >> 32),
			                      (unsigned int) entrylo0);
			printk("  %08x %08x", (unsigned int)(entrylo1 >> 32),
			                      (unsigned int) entrylo1);
/* {int i;for(i=100000000;i;i--);} */
		}
	}
	printk("\n");
}

#include <linux/kernel.h>
#include <linux/mm.h>

#include <asm/mipsconfig.h>
#include <asm/page.h>
#include <asm/pgtable.h>

extern volatile int buggy;

void
dump_list(void *address)
{
	pgd_t	*page_dir;
	pgd_t	*pgd;
	pmd_t	*pmd;
	pte_t	*pte;
	pte_t	page;
	unsigned int addr;
	unsigned int *pt;

	addr = (unsigned int) address;
	printk("Addr == %08x\n", addr);

	printk("current->tss.pg_dir == %08x\n", (unsigned int) current->tss.pg_dir);

	page_dir = pgd_offset(current, 0);
	printk("page_dir == %08x\n", (unsigned int) page_dir);

	pgd = pgd_offset(current, addr);
	printk("pgd == %08x, ", (unsigned int) pgd);

	pmd = pmd_offset(pgd, addr);
	printk("pmd == %08x, ", (unsigned int) pmd);

	pte = pte_offset(pmd, addr);
	printk("pte == %08x, ", (unsigned int) pte);

	page = *pte;
	printk("page == %08x\n", (unsigned int) pte_val(page));

	dump_tlb();
	printk("\n");

	pt = (unsigned int *) TLB_ROOT + (addr >> 22);
	printk("L1: *%08x == ", (unsigned int) pt);
	printk("%08x\n", *pt);

	pt = (unsigned int *) (TLBMAP + ((addr >> 10) & ~3));
	printk("L2: *%08x == ", (unsigned int) pt);
	printk("%08x\n", *pt);
}
